;;;;
;;;; Convert arbitrary precision floating point (represented using a
;;;; string) to a rational number and vice versa.
;;;;

(ns net.serice.clojure.rational-to-string
  (:gen-class))

(defn rational-to-string-version []
  "Return the version for the rational-to-string package."
  ["1.0.5" 1 0 5])

(defn- is-white-space [c]
  (some (partial = c) '(\space \tab \return \newline)))

(defn- skip-white-space [cs]
  "Return a vector holding the rest of the list of characters CS after
leading white space characters have be removed and the number of white
space characters that were removed:

    [others count]"
  (let [[white others] (split-with is-white-space cs)]
    [(if (empty? others) nil others) (count white)]))

(defn- is-decimal-digit [c]
  (some (partial = c) '(\0 \1 \2 \3 \4 \5 \6 \7 \8 \9)))

(defn- abs [x]
  (if (>= x 0) x (-' x)))

(defmulti ^:private floor
  "Return the the largest integer not greater than the rational number X.
For example (floor 1.5) = 1 and (floor -1.5) = -2."
  class)

(defmethod floor java.lang.Long [x] x)
(defmethod floor clojure.lang.BigInt [x] x)
(defmethod floor clojure.lang.Ratio [x]
  ;; Note that clojure.lang.Ratio types are guaranteed to have a
  ;; non-zero fractional part.  Clojure converts any
  ;; clojure.lang.Ratio that has a denominator of 1 (after reducing)
  ;; to a java.lang.Long or clojure.lang.BigInt.
  (-' (quot (.numerator x) (.denominator x))
     (if (neg? x) 1 0)))

(defmulti ^:private ceil
  "Return the the smallest integer not less than the rational number X.
For example (ceil 1.5) = 2 and (ceil -1.5) = -1."
  class)

(defmethod ceil java.lang.Long [x] x)
(defmethod ceil clojure.lang.BigInt [x] x)
(defmethod ceil clojure.lang.Ratio [x]
  ;; Note that clojure.lang.Ratio types are guaranteed to have a
  ;; non-zero fractional part.  Clojure converts any
  ;; clojure.lang.Ratio that has a denominator of 1 (after reducing)
  ;; to a java.lang.Long or clojure.lang.BigInt.
  (+' (quot (.numerator x) (.denominator x))
     (if (pos? x) 1 0)))

(defn- common-round [x]
  "Round the rational X to the nearest integer.  Ties are rounded to
the largest absolute value."
  (if (neg? x)
    (ceil  (-' x (/ 1 2)))
    (floor (+' x (/ 1 2)))))

(defn- expt [x y]
  "With some efficiency calculate x raised to the y where y is an
integer.  For some reason, this seems to be missing from
clojure.lang.BigInt.  If x is zero, a \"division by zero\"
ArithmeticException is thrown."
  (cond (neg? y) (/ (expt x (-' y)))
        (zero? y) 1
        (even? y) (let [z (expt x (/ y 2))] (*' z z))
        :else (*' x (expt x (-' y 1N)))))

(defn- ilog10 [n]
  "With some efficiency, return the (floor (log10 n)) if (log10 n) is
positive and the (ceil (log10 n)) if (log10 n) is negative.  This is
used by (normalize-rational) to get a good approximation of how far it
needs to shift in order to normalize N."
  (if (<= n 0)
    (throw (new RuntimeException (format "ilog10: domain error: %s" n)))
    (loop [n n, x 10, x' 1, count 1, count' 0, acc 0]
      (cond (and (>= n 1) (< n 10)) (+' count' acc)
            (< n 1) (- (ilog10 (/ n)))
            (>= n x) (recur n (*' x x) x (*' 2 count) count acc)
            :else (recur (/ n x') 10 1 1 0 (+' count' acc))))))

(defn approx-log10 [n]
    "Return the approximate log (base-10) of the rational number N
such that (1 <= (abs (expt n (approx-log10 n))) < 10).  This is used
by (normalize-rational) to normalize N."
    (letfn
        ;; Simple but very inefficient loop that increments or
        ;; decrements the exponent until n is normalized.
        ;; Fortunately, the approximation below that uses (ilog10) is
        ;; very good such that this loop should iterate only about twice.
        ((iter [x exponent]
           (let [y (expt 10 exponent)
                 z (/ x y)]
             (cond (>= z 10) (iter x (+' exponent 1))
                   (< z 1) (iter x (-' exponent 1))
                   :else exponent))))
      (if (<= n 0)
        (throw (new RuntimeException "approx-log10: domain error: n = ~a" n))
        (if (and (>= n 1) (< n 10))
          0
          ;; Get a very good approximation using (ilog10) and pass
          ;; it into (iter) which does a linear search from there.
          (iter n (ilog10 n))))))

(defn normalize-rational [n]
  "Normalize the non-negative, rational number N by returning the vector

    [SIGNIFICAND EXPONENT]

such that

    N = SIGNIFICAND * 10 ^ EXPONENT

where 1 <= SIGNIFICAND < 10."
  (if (zero? n)
    [n 0]
    (let [exponent (approx-log10 (abs n))]
      [(*' n (expt 10 (-' exponent))) exponent])))

(defn- make-zeros [n & acc]
  "Prepend N zero characters to acc and return as a string."
  (if (<= n 0) acc (recur (dec n) (cons \0 acc))))

(defn- add-zeros [s precision]
  "Add leading zeros to the scaled value held in the string S.
This is needed so that, when numbers like 0.001 are scaled (by
multiplying below by (expt 10 precision)), the significant zeros just
to the right of the decimal place are not lost. This is important
because we do the scaling as a cheap way of getting all of the digits
to the left-hand side of the decimal point where (str) can convert
them to a string.  After inserting all the digits into what we are
calling the scaled string S, it is still necessary to go back into S
and insert the decimal point which requires all significant zeros to
be present."
  (let [slen (count s)]
    (if (< slen precision)
      (.concat (clojure.string/join (make-zeros (-' precision slen))) s)
      s)))

(defn- trim-trailing-zeros [s]
  "Trim the trailing zero characters from the string S.  If the number
is of the form \"1.000\", the decimal point is also trimmed, but none
of the significant digits to the left of the decimal point even if
those digits are zeros.  If the number is of the form \"1000\", none
of the zeros are trimmed because they are all significant."
  ;; If s does not have a decimal point, all the digits are significant.
  (if (< (.indexOf s (int \.)) 0)
    s
    (loop [[c & cs :as s] (reverse s)]
      (cond
       ;; First base case.  Note that if we reach the base case and the
       ;; accumulator is still empty, it means all the digits were
       ;; trailing zeros and were removed in which case we have to force
       ;; #\0 to be returned.
       (empty? s) "0"
       ;; Second base case.  Any non-zero character means we have
       ;; reached the end of the trailing zeros.  You want to return
       ;; strings like "1" instead of "1." in this one case too so we
       ;; skip the trailing decimal point too.
       (not (= c \0))
       (clojure.string/join (reverse (if (= c \.) cs s)))
       ;; Must be a trailing zero.
       :else (recur cs)))))

(defn- insert-decimal-point [s precision]
  "Insert the decimal point to the left of the right-most PRECISION
digits in the string S."
  (let [pivot (-' (count s) precision)]
    (when (< pivot 0)
      (throw (new RuntimeException "Internal error.  Pivot too small.")))
    (.concat (if (== pivot 0) "0" (.substring s 0 pivot))
             (if (== precision 0) "" (.concat "." (.substring s pivot))))))

(defn- to-basic-string [n precision trim]
  "Convert the rational number N to the string representation of a
basic floating point number with PRECISION worth of digits after the
decimal point.  If TRIM is true, trailing zeros will be trimmed."
  (let [ x (common-round (*' (abs n) (expt 10 precision)))
         s (str x)
         s (add-zeros s precision)
         s (insert-decimal-point s precision)
        rv (if trim (trim-trailing-zeros s) s)]
    (if (< n 0) (.concat "-" rv) rv)))

(defn- to-sci-string [n precision trim estr]
  "Convert the rational number N to the string representation of a
floating point number in normalized scientific notation with PRECISION
worth of digits after the decimal point.  The value of the string ESTR
will be used to separate the significand from the exponent.
Typically, ESTR is \"e\" or \"E\".  If TRIM is true, trailing zeros
will be trimmed."
  (let [[x exponent] (normalize-rational n)]
    (reduce (fn [x y] (.concat x y))
            ""
            [(to-basic-string x precision trim) estr (str exponent)])))

(defn cs-to-sign [[sign & cs :as lst]]
  "Consume the head of the character list LST if it holds one of the
sign characters \\+ or \\- and return a list holding the sign, the
rest of the character list, and the number of characters consumed:

    [sign rest 1]

Otherwise, return [nil lst 0]."
  (cond (= sign \+) [:plus cs 1]
        (= sign \-) [:minus cs 1]
        :else [nil lst 0]))

(defn cs-to-uint [lst]
  "Convert the head of the character list LST to an unsigned integer
and return a list that holds the unsigned integer, the rest of the
character list, and the number of characters consumed.  If the head
cannot be converted to an unsigned integer, [nil lst 0] is returned.
Note this this function does NOT allow for a leading \\+ character.
This is because it is used for getting the fractional part (after the
decimal) which does not allow for a leading \\+ character.  If you
need this functionality use (cs-to-int) instead."
  (letfn
      ((to-bigint [cs]
         (cond (empty? cs) nil
               :else (bigint (clojure.string/join cs))))
       (iter [[c & cs :as lst] acc count]
         (cond (not (is-decimal-digit c)) [(to-bigint acc) lst count]
               :else (iter cs (conj acc c) (inc' count)))))
    (iter lst [] 0)))

(defn cs-to-int [lst]
  "Convert the head of the character list LST to a signed integer and
return a list that holds the signed integer, the rest of the character
list, and the number of characters consumed.  If the head cannot be
converted to a signed integer, [nil lst 0] is returned."
  (let [[sign-part  cs1 sign-count] (cs-to-sign lst)]
    (cond
     ;; Positive.
     (= sign-part :plus)
     (let [[x cs2 main-count] (cs-to-uint cs1)]
       (if x [x cs2 (+' sign-count main-count)] [nil lst 0]))
     ;; Negative.
     (= sign-part :minus)
     (let [[x cs2 main-count] (cs-to-uint cs1)]
       (if x [(-' x) cs2 (+' sign-count main-count)] [nil lst 0]))
     ;; Otherwise.
     :else (cs-to-uint lst))))

(defn cs-to-fract [[point & cs :as lst]]
  "Extract the fractional part from the the head of the character list
LST.  This function expects a leading \\. character.  If the
fractional part can be extracted

    [fract-part rest fract-count]

is returned; otherwise [nil lst 0] is returned.  Note that for numbers
like \"3.\" one character will be consumed even though there are no
digits in the fractional part."
  (if (not (= point \.))
    [nil lst 0]
    (let [[x cs count] (cs-to-uint cs)]
      [x cs (inc count)])))

(defn cs-to-expt [[emarker & cs :as lst]]
  "Extract the exponential part from the the head of the character
list LST.  This function expects a leading \\e or \\E character.  If
the exponential part can be extracted

    [expt-part rest expt-count]

is returned; otherwise [nil lst 0] is returned."
  (if (and (not (= emarker \e)) (not (= emarker \E)))
    [nil lst 0]
    (let [[x cs count] (cs-to-int cs)]
      (if x [x cs (inc count)] [nil lst 0]))))

(defn cs-to-decomposed-float [lst]
  "Convert the head of the character list LST to its decomposed
floating point representations and return a list that holds the
decomposed representation, the rest of the character list, and the
total number of characters consumed.  If the head can be converted,

    [[sign-part whole-part [fract-part fract-count] expt-part]
     rest
     total-count]

is returned where \"fract-count\" is the number of *digits* in the
fractional part.  This is needed so that you'll know how to covert
things like 0.3, 0.03, and 0.003 to the rational numbers (/ 3 10), (/
3 100), and (/ 3 1000) respectively.  \"total-count\" is the total
number of *characters* consumed.

If the head cannot be converted, [[nil nil nil nil] lst 0] is returned."

  (let [[sign-part  cs1 sign-count ] (cs-to-sign lst)
        [whole-part cs2 whole-count] (cs-to-uint cs1)
        [fract-part cs3 fract-count] (cs-to-fract cs2)
        [expt-part  cs4 expt-count ] (cs-to-expt cs3)]
    (if (and (not whole-part) (not fract-part))
      [[nil nil nil nil] lst 0]
      (let [total-count (+' sign-count whole-count fract-count expt-count)
            fract-digits (if fract-part (dec' fract-count) 0)]
        ;; To get the number of *digits* in the fractional part, we
        ;; have to subtract out 1 for the \. character.
        [[sign-part whole-part [fract-part fract-digits] expt-part]
         cs4 total-count]))))

(defn parse-decimal-string [s]
  "This function parses the string S which should be the string
representation of an arbitrary precision floating point number in
the the following form:

    \"[+|-]<whole_part>.<fract_part>[e|E[+|-]<exp_part>]\"

This function returns

    (values <ok>
            <sign>
            <whole_part>
            <fract_part>
            <fract_part_digit_count>
            <exp_part>
            <remainder_of_s>
            <total_consumed_character_count)

where <ok> is true if a decimal string could be extracted from the
front of S or false otherwise.  If <ok> is true, the trailing part of S
that did not contribute to the decimal string is returned as the
second to last value.

The <sign> value that is returned will be the symbol :plus or :minus
depending on whether the sign of the number is positive or negative.
<whole_part>, <fract_part>, and <exp_part> will hold the whole part,
fractional part, and exponential part respectively.

The <fract_count> is the number of decimal digits in <fract_part>.
This allows you calculate the fractional part using arbitrary
precision rational arithmetic as follows:

    (/ <fract_part> (expt 10 <fract_count>))"
  (let
      ;; Deconstruct (cs-to-decomposed-float) into into its component parts.
      [[[sign-part
          whole-part
          [fract-part fract-count]
          expt-part]
         rest
         total-count]
        (cs-to-decomposed-float s)]
    ;; Return the parsed decimal string.
    [(> total-count 0)
     sign-part
     whole-part
     fract-part
     fract-count
     expt-part
     rest
     total-count]))

(defn cs-to-rational [[c & cs :as lst]]
  "Convert the head of the character list LST to a rational and return
a list that holds the rational, the rest of the character list, and
the count of the number of characters consumed:

    [x rest count]

If the head cannot be converted to a rational, [nil lst 0] is
returned."
  ;; The first line destructures the second line.
  (let [[[sign-part whole-part [fract-part fract-count] expt-part] lst2 total]
        (cs-to-decomposed-float lst)]
    (if (and (not whole-part) (not fract-part))
      [nil lst 0]
      (let [whole-part (if whole-part whole-part 0)
            fract-part (if fract-part fract-part 0)
            fract-count (if fract-count fract-count 0)
            expt-part  (if expt-part  expt-part  0)
            magnitude (*' (+' whole-part (/ fract-part (expt 10N fract-count)))
                          (expt 10N expt-part))
            x (if (= sign-part :minus) (-' magnitude) magnitude)]
        [x lst2 total]))))

(defn string-to-rational [s]
  "Convert the head of the string S to a rational and return a list
that holds the rational, the rest of the string, and the count of the
number of characters consumed:

    [x rest count]

If the head cannot be converted to a rational, [nil s 0] is returned."
  (let
      ;; White space is only allowed at the front.  So, we strip it
      ;; off here.  All the other routines disallow white space.
      [[cs ws-count] (skip-white-space s)
       [x cs main-count] (cs-to-rational cs)]
    (if x
      [x (clojure.string/join cs) (+ ws-count main-count)]
      [nil s 0])))

(defn rational-to-string [n precision & {:keys [trim sci estr]
                                         :or {trim true sci nil estr "e"}}]
  "Convert the rational number N to a string holding the decimal
representation of N having PRECISION digits after the decimal point.
If TRIM is true, trailing zeros in the fractional part will be
trimmed; otherwise, trailing zeros will be returned.  If SCI is true,
the result will be in scientific notation with the value of ESTR
inserted between the significand and the exponent."
  (when (neg? precision)
    (throw (new RuntimeException "precision cannot be negative")))
  (if sci
    (to-sci-string n precision trim estr)
    (to-basic-string n precision trim)))
